"""
Implementation of the ``Weights``
"""
from typing import Any, cast

import numpy as np

import awesom.utilities as utils
from awesom.typing import FilePath, FloatArray, IntArray


class Weights:
    """Weights
    """
    def __init__(self, dx: int, dy: int, dw: int,
                 seed: int | IntArray | None = None) -> None:
        self.dx = dx
        self.dy = dy
        self.dw = dw
        self.shape = (self.dx, self.dy, self.dw)
        self.n_units = self.dx * self.dy
        self._vectors = np.empty((self.n_units, self.dw), dtype=np.float64)
        self._rng = np.random.default_rng(seed)


    def __getitem__(self, key: Any) -> FloatArray:
        return cast(FloatArray, self._vectors[key])


    @property
    def vectors(self) -> FloatArray:
        """Return weight vectors"""
        return self._vectors


    def update(self, buff: FloatArray) -> None:
        """Update the weight vectors

        Args:
            buff:   Weight updates
        """
        if buff.dtype != self._vectors.dtype:
            raise TypeError(f"Update buffer has incompatible type <{buff.dtype}>. "
                            f"Expected <{self._vectors.dtype}>.")

        if buff.shape != self._vectors.shape:
            raise TypeError(f"Update buffer has incompatible shape <{buff.shape}>. "
                            f"Expected <{self._vectors.shape}>")

        np.add(buff, self._vectors, out=self._vectors)


    def init_pca(self, training_data: FloatArray | None = None,
                 adapt: bool = True) -> None:
        """Initialize weights using PCA method

        Compute initial SOM weights by sampling from the first two principal
        components of the input data set.

        Args:
            trainig_data:  Input data set
            adapt:  If ``True``, the largest value of ``shape`` is applied to the
                    principal component with the largest sigular value. This
                    orients the map, such that map dimension with the most units
                    coincides with principal component with the largest variance.
        """
        if training_data is None:
            training_data = self._rng.integers(-100, 100, (300, self.dw)).astype(float)
        _, vects, trans_data = utils.pca(training_data, 2)

        if adapt:
            shape = tuple(sorted((self.dx, self.dy), reverse=True))
        else:
            shape = (self.dx, self.dy)

        data_min = trans_data.min(axis=0)
        data_max = trans_data.max(axis=0)
        dim_x = np.linspace(data_min[0], data_max[0], shape[0])
        dim_y = np.linspace(data_min[1], data_max[1], shape[1])

        grid_x, grid_y = np.meshgrid(dim_x, dim_y)
        points = np.vstack((grid_x.ravel(), grid_y.ravel()))
        self._vectors[...] = points.T @ vects + training_data.mean(axis=0)


    def init_rnd(self, training_data: FloatArray | None = None) -> None:
        """Compute initial SOM weights by sampling uniformly from the data space.

        Args:
            dims:  Dimensions of SOM
            data:  Input data set. If ``None``, sample from [-10, 10]
        """
        if training_data is not None:
            data_limits = np.column_stack((training_data.min(axis=0),
                                           training_data.max(axis=0)))
        else:
            data_limits = self._rng.integers(-10, 10, (self.dw, 2))
            data_limits.sort()
        weights = [self._rng.uniform(dmin, dmax, self.dx*self.dy)
                   for (dmin, dmax) in data_limits]
        self._vectors[...] = np.column_stack(weights)


    def init_stv(self) -> None:
        """Initialize with stochastic vectors
        """
        nvt = self.dx * self.dy
        self._vectors[...] = utils.sample_st_vector(nvt, self.dw)


    def init_stm(self) -> None:
        """Initialize with stochastic matrices
        """
        nvt = self.dx * self.dy
        self._vectors[...] = utils.sample_st_matrix(nvt, self.dw)


    def save_vectors(self, path: FilePath) -> None:
        """Store weight vector in a portable `.npy` file

        Args:
            path:  File path
        """
        np.save(path, self._vectors, allow_pickle=False)



"""
Grid for Self-organizing maps
"""
from typing import Generator

import numpy as np
from scipy.spatial import cKDTree

from awesom.typing import IntArray, DistFunc, Shape, FloatArray


class SomGrid:
    """Rectangular SOM grid
    """
    def __init__(self, shape: Shape) -> None:
        if not all(isinstance(val, int) and val >= 1 for val in shape):
            raise ValueError("Dimensions must be integer > 0.")
        self.shape = shape
        self.pos = np.asarray(list(np.ndindex(shape)), dtype=int)
        self.tree = cKDTree(self.pos)
        self.rows, self.cols = np.indices(shape)
        self.dists = np.empty((self.pos.shape[0], 1), dtype=np.float64)

    def nhb_idx(self, radius: float, points: IntArray | None = None) -> IntArray:
        """Compute the neighbourhood unit indices within ``radius``

        If ``points`` is given, return the neighbourhood around each unit in
        ``points``. Otherwise, return the neighbourhodd for each unit on the
        grid.

        Args:
            point:   Coordinate in a two-dimensional array.
            radius:  Lenght of radius.

        Returns:
            Array of indices of neighbours.
        """
        if points is None:
            points = self.pos
        return np.asarray(self.tree.query_ball_point(points, radius, np.inf))

    def nhb(self, radius: float, points: IntArray | None = None) -> IntArray:
        """Compute neighbourhood within ``radius``

        If ``points`` is given, return the neighbourhood around each unit in
        ``points``. Otherwise, return the neighbourhodd for each unit on the
        grid.

        Args:
            point:   Coordinate in a two-dimensional array.
            radius:  Lenght of radius.

        Returns:
            Array of positions of neighbours.
        """
        if points is None:
            points = self.pos
        idx = self.nhb_idx(radius, points)
        return self.pos[idx]

    def neighbourhood_distances(self, center: FloatArray, radius: float,
                                distance_func: DistFunc) -> FloatArray:
        """Compute distances given ``distance_func``"""
        return distance_func(self.pos, center, radius, self.dists)

    def __iter__(self) -> Generator[tuple[int, int], None, None]:
        for row, col in zip(self.rows.flat, self.cols.flat):
            yield row, col



"""
Function for generating test and illustration data sets.
"""

import numpy as np
from scipy import stats

from awesom.typing import IntArray, FloatArray


def norm_circle(n_classes: int, n_per_class: int, class_std: int,
                center: tuple[int, int] = (0, 0), radius: int = 5,
                seed: int | None = None) -> tuple[FloatArray, IntArray]:
    # pylint: disable = too-many-arguments
    """Generate ``n_per_class`` samples from ``n_classes`` bivariate normal
    distributions, each with standard deviation ``class_std``. The means
    are equidistantly placed on a circle with radius ``radius``.

    Args:
        n_classes:    Number of classes.
        n_per_class:  Number of samples in each class.
        class_std:    Standard deviation for every class.
        center:       Center of ther circle.
        radius:       Radius of the circle on which the means are placed.
        seed:         Set the random seed.

    Returns:
        Data set and target vector.
    """
    n_samples = n_classes * n_per_class
    ang = np.pi * np.linspace(0, 360, n_classes, endpoint=False) / 180
    xy_pos = np.stack((np.sin(ang), np.cos(ang)), axis=1)
    xy_pos *= radius + np.asarray(center)

    out = np.empty((n_samples, 2))
    for i, pos in enumerate(xy_pos):
        idx = slice(i*n_per_class, (i+1)*n_per_class)
        distr = stats.multivariate_normal(pos, np.sqrt(class_std), seed=seed)
        out[idx, :] = distr.rvs(n_per_class)
    return out, np.arange(n_samples) // n_per_class



"""
Plotting functions for SOMs.
"""

from typing import Any, Callable

import numpy as np
import numpy.typing as npt

from awesom.typing import FloatArray, IntArray, Axis, Axes3D
from . import utilities as utils
from . som import SomBase


def umatrix(ax: Axis, som: SomBase, outline: bool = False, **kwargs: Any
            ) -> None:
    """Plot the U-matrix.

    Args:
        ax:   Axis subplot.
        som:  SOM instance.

    Note:
        Figure aspect is set to "eqaul".
    """
    props = {
        'cmap': "terrain",
        'levels': 20}
    props.update(kwargs)
    _generic_contour(ax, som.umatrix(), outline, **props)


def umatrix3d(ax: Axes3D, som: SomBase, **kwargs: Any) -> None:
    """Plot the U-matrix in three dimensions.

    Args:
        ax:   Axis subplot.
        som:  SOM instance.

    Note:
        Figure aspect is set to "eqaul".
    """
    props = {
        'cmap': "terrain",
        }
    props.update(kwargs)
    ax.plot_surface(*np.mgrid[:som.dx, :som.dy], som.umatrix(), **props)


def component(ax: Axis, som: SomBase, comp: int, outline: bool = False,
              **kwargs: Any) -> None:
    """Plot a component plane.

    Args:
        ax:    Axis subplot.
        som:   SOM instance.
        comp:  Component number.
    """
    props = {
        'cmap': "magma",
        'levels': 20,}
    props.update(kwargs)
    _generic_contour(ax, som.weights[:, comp].reshape(som.shape), outline,
                     **props)


def label_target(ax: Axis, som: SomBase, data: FloatArray, target: IntArray, **kwargs: Any
                 ) -> None:
    """Add target labels for each bmu.

    Args:
        ax:      Axis subplot.
        som:     SOM instance.
        data:    Input data.
        target:  Target labels.
    """
    props = {
        'fontsize': 9,
        'ha': "left",
        'va': "bottom",
        }
    props.update(kwargs)

    bmu = som.match(data)
    bmu_xy = np.fliplr(np.atleast_2d(bmu)).T
    for x, y, t in zip(*bmu_xy, target):
        ax.text(x, y, t, fontdict=props)


def qerror(ax: Axis, som: SomBase, **kwargs: Any) -> None:
    """Plot quantization error."""
    props = {
        'lw': 3,
        'alpha': .8,
        }
    props.update(kwargs)
    ax.plot(som.quantization_error, **props)    # type: ignore


def cluster_by(ax: Axis, som: SomBase, data: FloatArray, target: IntArray, **kwargs: Any
               ) -> None:
    """Plot bmu colored by ``traget``.

    Args:
        ax:      Axis subplot.
        som:     SOM instance.
        data:    Input data.
        target:  Target labels.
    """
    props = {
            's': 50,
            'c': target,
            'marker': "o",
            }
    props.update(kwargs)
    bmu = som.match(data)
    bmu_xy = np.fliplr(np.atleast_2d(bmu)).T
    ax.scatter(*bmu_xy, **props)    # type: ignore


def hit_counts(ax: Axis, som: SomBase,
               transform: Callable[[IntArray], IntArray] | None = None
               , **kwargs: Any) -> None:
    """Plot the winner histogram.

    Each unit is colored according to the number of times it was bmu.

    Args:
        ax:    Axis subplot.
        som:   SOM instance.
        mode:  Choose either "linear", or "log".
    """
    props = {
        'interpolation': None,
        'origin': "lower",
        'cmap': "Greys",
        }
    props.update(kwargs)
    data = som.hit_counts.reshape(som.shape)
    if transform is not None:
        data = transform(data)
    ax.imshow(data, **props)    # type: ignore


def wire(ax: Axis, som: SomBase,
         unit_size: float | FloatArray = 100.0, line_width: float = 1.0,
         highlight: FloatArray | None = None, labels: bool = False,
         unit_color: str = 'k', **kwargs: Any) -> None:
    # pylint: disable = too-many-locals, too-many-arguments
    """Plot the weight vectors of a SOM with two-dimensional feature space.

    Neighbourhood relations are indicate by connecting lines.

    Args:
        ax:          The axis subplot.
        som:         SOM instance.
        unit_size:   Size for each unit.
        line_width:  Width of the wire lines.
        highlight:   Index of units to be marked in different color.
        labels:      If ``True``, attach a box with coordinates to each unit.

    Returns:
        vlines, hlines, bgmarker, umarker
    """
    if isinstance(unit_size, np.ndarray):
        marker_size = utils.scale(unit_size, 10, 110)
    elif isinstance(unit_size, (int, float)):
        marker_size = np.repeat(unit_size, som.n_units)
    else:
        msg = ("Argument of parameter ``unit_size`` must be real scalar "
               "or one-dimensional numpy array.")
        raise ValueError(msg)
    marker_size_bg = marker_size + marker_size / 100 * 30

    bg_color: npt.ArrayLike = "w"
    hl_color: str = "r"

    line_props = {
        'color': "k",
        'alpha': 0.7,
        'lw': line_width,
        'zorder': 9,
        }
    line_props.update(kwargs)

    marker_bg_props = {
        's': marker_size_bg,
        'c': bg_color,
        'edgecolors': None,
        'zorder': 11,
        }

    marker_hl_props = {
        's': marker_size,
        'c': unit_color,
        'alpha': line_props["alpha"],
        'edgecolor': "None",
        'zorder': 12
        }

    if highlight is not None:
        bg_color = np.where(highlight, hl_color, bg_color)

    rsw = som.weights.reshape(*som.shape, 2)
    v_wx, v_wy = rsw.T
    h_wx, h_wy = np.rollaxis(rsw, 1).T
    _vlines = ax.plot(v_wx, v_wy, **line_props)    # type: ignore
    _hlines = ax.plot(h_wx, h_wy, **line_props)    # type: ignore
    _bgmarker = ax.scatter(v_wx, v_wy, **marker_bg_props)    # type: ignore
    _umarker = ax.scatter(v_wx, v_wy, **marker_hl_props)     # type: ignore

    font = {'fontsize': 4,
            'va': "bottom",
            'ha': "center",
            }

    bbox = {'alpha': 0.7,
            'boxstyle': "round",
            'edgecolor': "#aaaaaa",
            'facecolor': "#dddddd",
            'linewidth': .5,
            }

    if labels is True:
        for (sw_x, sw_y), (ix, iy) in zip(som.weights, np.ndindex(som.shape)):
            ax.text(sw_x+1.3, sw_y, f"({ix}, {iy})", font, bbox=bbox, zorder=13)
    ax.set_aspect("equal")


def data_2d(ax: Axis, data: FloatArray, colors: FloatArray, **kwargs: Any) -> None:
    """Scatter plot a data set with two-dimensional feature space.

    This just the usual scatter command with some reasonable defaults.

    Args:
        ax:      The axis subplot.
        data:    The data set.
        colors:  Colors for each elemet in ``data``.

    Returns:
        PathCollection.
    """
    props = {
        'alpha': 0.2,
        'c': colors,
        'cmap': "plasma",
        'edgecolors': "None",
        's': 10}
    props.update(kwargs)
    _ = ax.scatter(*data.T, **props)    # type: ignore


def _generic_contour(ax: Axis, data: FloatArray, outline: bool = False,
                     **kwargs: Any) -> None:
    """Contour plot.

    Args:
        ax:    Axis subplot.
        data:  Two-dimensional array.
    """
    sdx, sdy = data.shape
    overwrites = {
        'extent': (-0.5, sdy-0.5, -0.5, sdx-0.5),
        }
    kwargs.update(overwrites)
    _ = ax.contourf(data, **kwargs)
    _ = ax.set_xticks(range(sdy))
    _ = ax.set_yticks(range(sdx))
    if outline:
        ax.contour(data, cmap="Greys_r", alpha=.7)
    ax.set_aspect("equal")



"""
Self-organizing map base classes
"""

# pylint: disable = R0904, R0902, R0913

import pathlib
import pickle

import numpy as np
import numpy.typing as npt
from scipy.spatial import distance

from awesom.exceptions import NotCalibratedError
from awesom.typing import FloatArray, IntArray, Metric, Shape, SomDims, FilePath
from . import defaults
from . import grid
from . import neighbors
from . import utilities as utils
from . weights import Weights


class SomBase:
    """Self-organizing map base class
    """
    def __init__(self, dims: SomDims, n_iter: int, eta: float, nhr: float,
                 nh_shape: str, weights: Weights, metric: Metric = "euclidean",
                 seed: int | IntArray | None = None):

        self._grid = grid.SomGrid(dims[:2])
        self.n_features = dims[2]
        self._hit_counts = np.zeros(self.n_units, dtype=np.int_)
        self.n_iter = n_iter
        self.metric = metric
        self._qrr = np.zeros(n_iter)
        self._trr = np.zeros(n_iter)
        self._weights = weights
        self._rng = np.random.default_rng(seed)
        self._calibration: IntArray | None = None

        try:
            self._neighbourhood = getattr(neighbors, nh_shape)
        except AttributeError as err:
            raise AttributeError(f"Neighborhood shape {nh_shape} is unknown. "
                                 "Use one of `gaussian`, `mexican`, `rect`, or "
                                 "`star`") from err

        if 0 < eta <= 1.:
            self.init_eta = eta
        else:
            raise ValueError(f"Parameter ``eta``={self.init_eta} not in"
                             "range [0, 1]")

        if nhr >= 1:
            self.init_nhr = nhr
        else:
            raise ValueError("Neighbourhood radius must be int > 0.")


    @property
    def dims(self) -> SomDims:
        """Return the SOM dimensions."""
        return (*self._grid.shape, self.n_features)

    @property
    def dx(self) -> int:
        """Return the number of units along the first dimension."""
        return self._grid.shape[0]

    @property
    def dy(self) -> int:
        """Return the number of units along the second dimension."""
        return self._grid.shape[1]

    @property
    def dw(self) -> int:
        """Return the dimension of the weight vectors."""
        return self.n_features

    @property
    def n_units(self) -> int:
        """Return the total number of units on the SOM."""
        return self.dx * self.dy

    @property
    def shape(self) -> Shape:
        """Return the map shape."""
        return self._grid.shape

    @property
    def grid(self) -> grid.SomGrid:
        """Return the grid."""
        return self._grid

    @property
    def weights(self) -> FloatArray:
        """Return the weight vectors."""
        return self._weights.vectors

    @property
    def hit_counts(self) -> IntArray:
        """Return total hit counts for each SOM unit."""
        return self._hit_counts

    @property
    def quantization_error(self) -> FloatArray:
        """Return quantization error."""
        return self._qrr

    @property
    def topographic_error(self) -> FloatArray:
        """Return topographic error."""
        return self._trr

    def calibrate(self, data: FloatArray, target: IntArray) -> None:
        """Retrieve the target value of the best matching input data vector
        for each unit weight vector.

        Args:
            data:     Input data set.
            target:  Target labels.
        """
        bm_dv, _ = utils.best_match(data, self.weights, self.metric)
        self._calibration = target[bm_dv]

    def distribute(self, data: FloatArray) -> dict[int, list[int]]:
        """Distribute the vectors of ``data`` on the SOM.

        Indices of vectors n ``data`` are mapped to the index of
        their best matching unit.

        Args:
            data:  Input data set.

        Returns:
            Dictionary with SOM unit indices as keys. Each key maps to a list
            that holds the indices of rows in ``data``, which best match this
            key.
        """
        return utils.distribute(self.match(data), self.n_units)

    def match_flat(self, data: FloatArray) -> IntArray:
        """Return the index of the best matching unit for each vector in
        ``data``.

        Args:
            data:  Input data set.

        Returns:
            Array of SOM unit indices.
        """
        bmu, _ = utils.best_match(self.weights, data, self.metric)
        return bmu

    def match(self, data: FloatArray) -> IntArray:
        """Return the multi index of the best matching unit for each vector in
        ``data``.

        Caution: This function returns the multi index into the array.

        Args:
            data:  Input data set.

        Returns:
            Array of SOM unit indices.
        """
        bmu = self.match_flat(data)
        return np.column_stack(np.unravel_index(bmu, self.shape))

    def predict(self, data: FloatArray) -> IntArray:
        """Predict the SOM index of the best matching unit
        for each item in ``data``.

        Args:
            data:  Input data. Rows are items, columns are features.

        Returns:
            One-dimensional array of indices.
        """
        if self._calibration is None:
            raise NotCalibratedError("Cannot predict fromm not calibrated SOM")
        bmi, _ = utils.best_match(self.weights, data, self.metric)
        return self._calibration[bmi]

    def save(self, path: FilePath) -> None:
        """Save SOM object to pickle file

        Args:
            path: Save SOM to this path.
        """
        path = pathlib.Path(path)
        with path.open("wb") as file:
            pickle.dump(self, file)

    def transform(self, data: FloatArray) -> FloatArray:
        """Transform each item in ``data`` to feature space.

        This, in principle, returns best matching unit's weight vectors.

        Args:
            data:  Input data. Rows are items, columns are features.

        Returns:
            Position of each data item in the feature space.
        """
        bmi = self.match_flat(data)
        return self._weights[bmi]

    def umatrix(self, radius: int = 1, scale: bool = True, norm: bool = True
                ) -> FloatArray:
        """Compute U-matrix of SOM instance.

        Args:
            radius:   Map neighbourhood radius.
            scale:    If ``True``, scale each U-height by the number of the
                      associated unit's neighbours.
            norm:     Normalize U-matrix if ``True``.

        Returns:
            Unified distance matrix.
        """
        u_height = np.empty(self.n_units, dtype="float64")
        nhd_per_unit = self._grid.nhb_idx(radius)
        for i, nhd_idx in enumerate(nhd_per_unit):
            cwv = self._weights[[i]]
            nhd = self._weights[nhd_idx]
            u_height[i] = distance.cdist(cwv, nhd, self.metric).sum()
            if scale:
                u_height[i] /= len(nhd_idx)
        if norm:
            umax = u_height.max()
            if umax == 0:
                u_height = np.zeros_like(u_height)
            else:
                u_height /= u_height.max()
        return u_height.reshape(self.shape)


class BatchMap(SomBase):
    """Self-organizing map with batch training

    The batch training updates the weight vectors once for all input vectors.
    """
    def __init__(self, dims: SomDims, n_iter: int, eta: float, nhr: float,
                 nh_shape: str = "gaussian", weights: Weights | None = None,
                 metric: Metric = "euclidean", seed: int | None = None):

        if weights is None:
            weights = Weights(*dims)
            weights.init_pca()

        super().__init__(dims, n_iter, eta, nhr, nh_shape, weights, metric,
                         seed=seed)


class IncrementalMap(SomBase):
    """Self-organizing map with online learning algorithm

    The incremental, or online training updates the weight vectors for each
    input vector.
    """
    def __init__(self, dims: SomDims, n_iter: int, eta: float, nhr: float,
                 nh_shape: str = "gaussian", weights: Weights | None = None,
                 metric: Metric = "euclidean", seed: int | None = None):

        if weights is None:
            weights = Weights(*dims)
            weights.init_pca()

        super().__init__(dims, n_iter, eta, nhr, nh_shape, weights, metric,
                         seed=seed)

    def fit(self, train_data: FloatArray, target: IntArray | None = None,
            verbose: bool = False) -> None:
        """Fit the SOM to the ``training_data``"""
        eta_ = utils.decrease_linear(self.init_eta, self.n_iter, defaults.FINAL_ETA)
        nhr_ = utils.decrease_expo(self.init_nhr, self.n_iter, defaults.FINAL_NHR)

        _update_buffer = np.empty_like(self.weights)
        for (c_iter, c_eta, c_nhr) in zip(range(self.n_iter), eta_, nhr_):
            if verbose:
                print(f"iter: {c_iter:2} -- eta: {np.round(c_eta, 4):<5} -- "
                      f"nh: {np.round(c_nhr, 5):<6}")

            for fvect in self._rng.permutation(train_data):
                bmu, err = utils.best_match(self.weights, fvect, self.metric)
                self._hit_counts[bmu] += 1
                m_idx = np.atleast_2d(np.unravel_index(bmu, self.shape)).T
                self._compute_update(fvect, m_idx, c_nhr, c_eta, _update_buffer)
                self._weights.update(_update_buffer)

            _, err = utils.best_match(self.weights, train_data, self.metric)
            self._qrr[c_iter] = err.sum() / train_data.shape[0]

        if target is not None:
            self.calibrate(train_data, target)

    def _compute_update(self, vec: FloatArray, center: npt.ArrayLike, radius:
                        float, eta: float, buffer: FloatArray) -> None:

        center = np.asarray(center).astype(np.float64)
        self.grid.neighbourhood_distances(center, radius, self._neighbourhood)
        np.subtract(vec, self._weights.vectors, out=buffer)
        np.multiply(self.grid.dists, buffer, out=buffer)
        np.multiply(eta, buffer, out=buffer)



"""
Utilities for self.organizing maps.
"""

import itertools
from typing import cast, Iterable, Iterator

import numpy as np
from scipy.spatial import distance
from scipy import stats

from awesom.typing import IntArray, FloatArray, Metric


def grid_iter(n_rows: int, n_cols: int) -> Iterator[tuple[int, int]]:
    """Compute grid indices of an two-dimensional array.

    Args:
        n_rows:  Number of array rows.
        n_cols:  Number of array columns.

    Returns:
        Multi-index iterator.
    """
    return itertools.product(range(n_rows), range(n_cols))


def grid(n_rows: int, n_cols: int) -> IntArray:
    """Compute grid indices of a two-dimensional array.

    Args:
        n_rows:  Number of array rows.
        n_cols:  Number of array columns.

    Returns:
        Two-dimensional array in which each row represents an multi-index.
    """
    return np.array(list(grid_iter(n_rows, n_cols)), dtype=int)


def decrease_linear(start: float, step: float, stop: float = 1.0
                    ) -> Iterator[float]:
    """Linearily decrease ``start``  in ``step`` steps to ``stop``."""
    if step < 1 or not isinstance(step, int):
        raise ValueError("Param `step` must be int >= 1.")
    if step == 1:
        yield start
    else:
        coef = (stop - start) / (step - 1)
        for stp in range(step):
            yield coef * stp + start


def decrease_expo(start: float, step: float, stop: float = 1.0
                  ) -> Iterator[float]:
    """Exponentially decrease ``start``  in ``step`` steps to ``stop``."""
    if step < 1 or not isinstance(step, int):
        raise ValueError("Param `step` must be int >= 1.")
    if step == 1:
        yield start
    else:
        coef = np.log(stop / start) / (step - 1)
        for stp in range(step):
            yield start * np.exp(coef*stp)


def best_match(weights: FloatArray, inp: FloatArray, metric: Metric,
               out: FloatArray | None = None) -> tuple[IntArray, FloatArray]:
    """Compute the best matching unit of ``weights`` for each
    element in ``inp``.

    If several elemets in ``weights`` have the same distance to the
    current element of ``inp``, the first element of ``weights`` is
    choosen to be the best matching unit.

    Args:
        weights:    Two-dimensional array of weights, in which each row
                    represents an unit.
        inp:        Array of test vectors. If two-dimensional, rows are
                    assumed to represent observations.
        metric:     Distance metric to use.

    Returns:
        Index and error of best matching units.
    """
    if weights.ndim != 2:
        msg = (f"Array ``weights`` has {weights.ndim} dimensions, it "
               "has to have exactly two dimensions.")
        raise ValueError(msg)

    if weights.shape[-1] != inp.shape[-1]:
        msg = (f"Feature dimension of ``weights`` has {weights.shape[0]} "
               "elemets, whereas ``inp`` has {inp.shape[-1]} elemets. "
               "However, both dimensions have to match exactly.")
        raise ValueError(msg)

    inp = np.atleast_2d(inp)
    if inp.ndim > 2:
        msg = (f"Array ``inp`` has {weights.ndim} dimensions, it "
               "has to have one or two dimensions.")
        raise ValueError(msg)
    dists = distance.cdist(weights, inp, metric, out=out)
    return dists.argmin(axis=0), dists.min(axis=0)


def sample_st_matrix(n_mat: int, size: int) -> FloatArray:
    """Sample stochastic matrices from Dirichlet distribution

    The distribution is configured to place five times more probability mass on
    the main diagonal than on the remaining elements.

    Args:
        n_mat:  Number of matrices
        size:   Number of matirx row/cols

    Returns:
        Two-dimensional array. Each row corresponds to a flattened matrix.
    """
    if n_mat < 1:
        raise ValueError("n_mat < 1")

    if size < 2:
        raise ValueError("size < 2")

    pfact = 5.0
    alpha = np.ones((size, size), dtype=np.float64)
    np.fill_diagonal(alpha, pfact)

    samples = [stats.dirichlet(a).rvs(n_mat) for a in alpha]
    return np.hstack(samples, dtype=np.float64)


def sample_st_vector(n_vectors: int, size: int) -> FloatArray:
    """Sample stochastic vectors

    Sample random stochastic vectors with uniformly distributed probability
    mass. The sum of each vector equals 1.0 and each element is a number
    between 0.0 and 1.0.

    Args:
        n_vectors: Number of vectors
        size:      Vector size

    Returns:
        Two-dimensional array, whose rows correspond to vectors
    """
    if n_vectors < 1:
        raise ValueError("``n_vectors < 0")

    if size < 1:
        raise ValueError("``size`` < 1")

    alpha = np.ones(size, dtype=np.float64)
    samples = stats.dirichlet(alpha).rvs(n_vectors)
    return np.asanyarray(samples, dtype=np.float64)


def distribute(bmu_idx: Iterable[int], n_units: int
               ) -> dict[int, list[int]]:
    """List training data matches per SOM unit.

    This method assumes that the ith element of ``bmu_idx`` corresponds to the
    ith vetor in a array of input data vectors.

    Empty units result in empty list.

    Args:
        bmu_idx:  Indices of best matching units.
        n_units:  Number of units on the SOM.

    Returns:
        Dictionary in which the keys represent the flat indices of SOM units.
        The corresponding value is a list of indices of those training data
        vectors that have been mapped to this unit.
    """
    unit_matches: dict[int, list[int]] = {i: [] for i in range(n_units)}
    for data_idx, bmu in enumerate(bmu_idx):
        unit_matches[bmu].append(data_idx)
    return unit_matches


def pca(data: FloatArray, n_comps: int = 2
        ) -> tuple[FloatArray, FloatArray, FloatArray]:
    """Perfom principal component analysis

    Interanlly, ``data`` will be centered but not scaled.

    Args:
        data:     Data set
        n_comps:  Number of principal components

    Returns:
        ``n_comps`` largest singular values,
        ``n_comps`` largest eigen vectors,
        transformed input data.
    """
    data_centered = data - data.mean(axis=0)
    _, vals, vects = np.linalg.svd(data_centered)

    ord_idx = np.flip(vals.argsort())[:n_comps]
    vals = vals[ord_idx]
    vects = vects[ord_idx]
    return vals, vects, data_centered @ vects.T


def scale(arr: FloatArray, new_min: int = 0, new_max: int = 1, axis: int = -1
          ) -> FloatArray:
    """Scale ``arr`` between ``new_min`` and ``new_max``

    Args:
        arr:        Array to be scaled.
        new_min:    Lower bound.
        new_max:    Upper bound.

    Return:
        One-dimensional array of transformed values.
    """
    xmax = arr.max(axis=axis, keepdims=True)
    xmin = arr.min(axis=axis, keepdims=True)

    fact = (arr-xmin) / (xmax - xmin)
    out = fact * (new_max - new_min) + new_min

    return cast(FloatArray, out)



"""
Neighborhood computations
"""
from typing import cast
import numpy as np
import numpy.typing as npt

from scipy.spatial import distance

from awesom.typing import IntArray, FloatArray, Shape, Coord


def gaussian(grid: IntArray, center: npt.ArrayLike, radius: float,
             out: FloatArray | None = None) -> FloatArray:
    """Compute n-dimensional Gaussian neighbourhood.

    Gaussian neighborhood smoothes the array.

    Params:
        grid      Array of n-dimensional indices.
        center    Index of the neighborhood center.
        radius    Size of neighborhood.
    """
    if radius <= 0:
        raise ValueError("Radius <= 0")

    if out is None:
        out = np.empty((grid.shape[0], 1), dtype=np.float64)

    distance.cdist(grid, center, metric="sqeuclidean", out=out)
    np.multiply(out, -1, out=out)
    np.divide(out, 2*radius**2, out)
    np.exp(out, out=out)
    return out


def mexican(grid: IntArray, center: npt.ArrayLike, radius: float
            ) -> FloatArray:
    """Compute n-dimensional Mexcican hat neighbourhood.

    Mexican hat neighborhood smoothes the array.

    Params:
        grid      Array of n-dimensional indices.
        center    Index of the neighborhood center.
        radius    Size of neighborhood.
    """
    if radius <= 0:
        raise ValueError("Radius <= 0")

    rsq = radius**2
    dists = np.empty((grid.shape[0], 1), dtype=np.float64)
    gauss = np.empty((grid.shape[0], 1), dtype=np.float64)
    norm = np.empty((grid.shape[0], 1), dtype=np.float64)

    distance.cdist(grid, center, metric="sqeuclidean", out=dists)

    np.divide(-dists, rsq, out=norm)
    np.divide(norm, 2, out=gauss)
    np.add(1, norm, out=norm)
    np.exp(gauss, out=dists)
    np.multiply(norm, dists, out=dists)
    return dists


def star(grid: IntArray, center: npt.ArrayLike, radius: float) -> FloatArray:
    """Compute n-dimensional cityblock neighborhood.

    The cityblock neighborhood is a star-shaped area
    around ``center``.

    Params:
        grid      Array of n-dimensional indices.
        center    Index of the neighborhood center.
        radius    Size of neighborhood.

    Returns:
    """
    dists = np.empty((grid.shape[0], 1), dtype=np.float64)
    distance.cdist(grid, center, metric="cityblock", out=dists)
    return np.less_equal(dists, radius, out=dists)


def neighborhood(grid: IntArray, metric: str = "sqeuclidean") -> FloatArray:
    """Compute n-dimensional cityblock neighborhood.

    The cityblock neighborhood is a star-shaped area
    around ``center``.

    Params:
        grid:      Array of n-dimensional indices.
        metric:    Distance metric.

    Returns:
        Pairwise distances of map units.
    """
    dists = distance.pdist(grid, metric)
    return cast(npt.NDArray[np.float64], distance.squareform(dists))


def rect(grid: IntArray, center: npt.ArrayLike, radius: float) -> FloatArray:
    """Compute n-dimensional Chebychev neighborhood.

    The Chebychev neighborhood is a square-shaped area
    around ``center``.

    Params:
        grid      Array of n-dimensional indices.
        center    Index of the neighborhood center.
        radius    Size of neighborhood.

    Returns:
        Two-dimensional array of in
    """
    dists = np.empty((grid.shape[0], 1), dtype=np.float64)
    distance.cdist(grid, center, metric="chebychev", out=dists)
    return np.less_equal(dists, radius)


def check_bounds(shape: Shape, point: Coord) -> bool:
    """Return ``True`` if ``point`` is valid index in ``shape``.

    Args:
        shape:  Shape of two-dimensional array.
        point:  Two-dimensional coordinate.

    Return:
        True if ``point`` is within ``shape`` else ``False``.
    """
    return (0 <= point[0] < shape[0]) and (0 <= point[1] < shape[1])


def direct_rect_nb(shape: Shape, point: Coord) -> IntArray:
    """Return the set of direct neighbours of ``point`` given rectangular
    topology.

    Args:
        shape:  Shape of two-dimensional array.
        point:  Two-dimensional coordinate.

    Returns:
        Advanced index of points in neighbourhood set.
    """
    nhb = []
    for i in range(point[0]-1, point[0]+2):
        for j in range(point[1]-1, point[1]+2):
            if check_bounds(shape, (i, j)):
                nhb.append((i, j))
    return np.asarray(nhb)



"""
Type aliases
"""
import pathlib
from typing import Any, Callable

from matplotlib import axes
from mpl_toolkits import mplot3d
import numpy as np
import numpy.typing as npt


IntArray = np.ndarray[Any, np.dtype[np.int_]]
FloatArray = np.ndarray[Any, np.dtype[np.float64]]

Axis = axes.Axes
Axes3D = mplot3d.axes3d.Axes3D

Coord = tuple[int, int]
Shape = tuple[int, int]
SomDims = tuple[int, int, int]

Metric = str | Callable[[FloatArray, FloatArray], float]

FilePath = pathlib.Path | str

DistFunc = Callable[
    [IntArray, npt.ArrayLike, float, FloatArray | None],
    FloatArray]
